@@ -1,8 +1,8 @@
 /*
  * curl.c
- * $Id$
  *
- * Copyright (c) 2005 Paul Guyot, The MacPorts Project.
+ * Copyright (c) 2005 Paul Guyot
+ * Copyright 2006-2011, 2013-2014 The MacPorts Project
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -41,6 +41,8 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <strings.h>
+#include <sys/select.h>
 #include <utime.h>
 
 #include <curl/curl.h>
@@ -58,25 +60,42 @@
 #define _CURL_MINIMUM_XFER_TIMEOUT	((long)(60))		/* 1 minute */
 #define _CURL_MINIMUM_PROGRESS_INTERVAL ((double)(0.2)) /* 0.2 seconds */
 
+#if defined CURLOPT_ACCEPT_ENCODING
+#define _CURL_ENCODING CURLOPT_ACCEPT_ENCODING
+#else
+#define _CURL_ENCODING CURLOPT_ENCODING
+#endif
+
 /* ========================================================================= **
  * Definitions
  * ========================================================================= */
 
 /* ------------------------------------------------------------------------- **
- * Global cURL handle
+ * Global curl handles
  * ------------------------------------------------------------------------- */
-/* we use a single global handle rather than creating and destroying handles to
+/* If we want to use TclX' signal handling mechanism we need curl to return
+ * control to our code from time to time so we can call Tcl_AsyncInvoke to
+ * process pending signals. To do that, we could either abuse the curl progress
+ * callback (which would mean we could no longer use the default curl progress
+ * callback, or we need to use the curl multi API. */
+static CURLM* theMHandle = NULL;
+/* We use a single global handle rather than creating and destroying handles to
  * take advantage of HTTP pipelining, especially to the packages servers. */
 static CURL* theHandle = NULL;
 
+/* Global libcurl version info. */
+static curl_version_info_data *libcurl_version_info = NULL;
+
 /* ------------------------------------------------------------------------- **
  * Prototypes
  * ------------------------------------------------------------------------- */
 int SetResultFromCurlErrorCode(Tcl_Interp* interp, CURLcode inErrorCode);
-int CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
-int CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
-int CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
-int CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[]);
+int SetResultFromCurlMErrorCode(Tcl_Interp* interp, CURLMcode inErrorCode);
+int CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
+int CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
+int CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
+int CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
+int CurlVersionCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[]);
 
 typedef struct {
 	Tcl_Interp *interp;
@@ -88,6 +107,7 @@ static int CurlProgressHandler(tcl_callback_t *callback, double dltotal, double
 static void CurlProgressCleanup(tcl_callback_t *callback);
 
 void CurlInit(void);
+static void set_curl_version_info(void);
 
 /* ========================================================================= **
  * Entry points
@@ -116,23 +136,107 @@ SetResultFromCurlErrorCode(Tcl_Interp *interp, CURLcode inErrorCode)
 	return result;
 }
 
+/**
+ * Set the result if a libcurl multi error occurred return TCL_ERROR.
+ * Otherwise, set the result to "" and return TCL_OK.
+ *
+ * @param interp		pointer to the interpreter.
+ * @param inErrorCode	code of the multi error.
+ * @return TCL_OK if inErrorCode is 0, TCL_ERROR otherwise.
+ */
+int
+SetResultFromCurlMErrorCode(Tcl_Interp *interp, CURLMcode inErrorCode)
+{
+	int result = TCL_ERROR;
+
+	if (inErrorCode == CURLM_OK) {
+		Tcl_SetResult(interp, "", TCL_STATIC);
+		result = TCL_OK;
+	} else {
+		Tcl_SetResult(interp, (char *)curl_multi_strerror(inErrorCode), TCL_VOLATILE);
+	}
+
+	return result;
+}
+
+/* Bug affecting some versions of Monterey through Sequoia:
+   reusing an easy handle for different protocols can cause a
+   crash. https://github.com/curl/curl/issues/13731
+   TODO: Add upper bound here when we know what version macOS 16 ships. */
+#if LIBCURL_VERSION_NUM >= 0x074d00
+static void cleanup_handle_if_needed(const char *theURL);
+
+static void cleanup_handle_if_needed(const char *theURL)
+{
+    static CURLU *currentURLHandle = NULL;
+    static CURLU *previousURLHandle = NULL;
+
+    set_curl_version_info();
+    if (libcurl_version_info == NULL) {
+        fprintf(stderr, "Warning: set_curl_version_info failed\n");
+        return;
+    }
+
+    if (libcurl_version_info->version_num >= 0x080600 && libcurl_version_info->version_num < 0x080800) {
+        if (previousURLHandle != NULL) {
+            curl_url_cleanup(previousURLHandle);
+        }
+        previousURLHandle = currentURLHandle;
+        currentURLHandle = curl_url();
+         if (currentURLHandle != NULL
+            && curl_url_set(currentURLHandle, CURLUPART_URL, theURL, 0) != CURLUE_OK) {
+            curl_url_cleanup(currentURLHandle);
+            currentURLHandle = NULL;
+        }
+        if (theMHandle != NULL && currentURLHandle != NULL && previousURLHandle != NULL) {
+            char *current_scheme = NULL;
+            char *previous_scheme = NULL;
+            char *current_host = NULL;
+            char *previous_host = NULL;
+            int must_recreate_handle = 0;
+            if (curl_url_get(previousURLHandle, CURLUPART_SCHEME, &previous_scheme, 0) == CURLUE_OK
+                && curl_url_get(currentURLHandle, CURLUPART_SCHEME, &current_scheme, 0) == CURLUE_OK
+                && 0 != strcasecmp(previous_scheme, current_scheme)) {
+                must_recreate_handle = 1;
+            }
+            if (!must_recreate_handle
+                && curl_url_get(previousURLHandle, CURLUPART_HOST, &previous_host, 0) == CURLUE_OK
+                && curl_url_get(currentURLHandle, CURLUPART_HOST, &current_host, 0) == CURLUE_OK
+                && 0 != strcasecmp(previous_host, current_host)) {
+                must_recreate_handle = 1;
+            }
+            if (must_recreate_handle) {
+                /* deallocate the handle and start again */
+                curl_multi_cleanup(theMHandle);
+                theMHandle = NULL;
+            }
+        }
+    }
+}
+#endif /* LIBCURL_VERSION_NUM >= 0x074d00 */
+
 /**
  * curl fetch subcommand entry point.
  *
- * syntax: curl fetch [--disable-epsv] [--ignore-ssl-cert] [--remote-time] [-u userpass] [--effective-url lasturlvar] [--progress "builtin"|callback] url filename
+ * syntax: curl fetch [--disable-epsv] [--ignore-ssl-cert] [--remote-time] [-u userpass] [--effective-url lasturlvar] [--progress "builtin"|callback] [--enable-compression] url filename
  *
  * @param interp		current interpreter
  * @param objc			number of parameters
  * @param objv			parameters
  */
 int
-CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
+	bool handleAdded = false;
 	FILE* theFile = NULL;
-	bool performFailed = false;
 	char theErrorString[CURL_ERROR_SIZE];
 
+	/* Always 0-initialize the error string, since older curl versions may not
+	 * initialize the error string buffer at all. See
+	 * https://trac.macports.org/ticket/60581. */
+	theErrorString[0] = '\0';
+
 	do {
 		int noprogress = 1;
 		int useepsv = 1;
@@ -147,13 +251,20 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		};
 		char* effectiveURL = NULL;
 		char* userAgent = PACKAGE_NAME "/" PACKAGE_VERSION " libcurl/" LIBCURL_VERSION;
+		const int MAXHTTPHEADERS = 100;
+		int numHTTPHeaders = 0;
+		const char* httpHeaders[MAXHTTPHEADERS];
 		int optioncrsr;
 		int lastoption;
 		const char* theURL;
 		const char* theFilePath;
 		long theFileTime = 0;
 		CURLcode theCurlCode;
+		CURLMcode theCurlMCode;
 		struct curl_slist *headers = NULL;
+		struct CURLMsg *info = NULL;
+		int running; /* number of running transfers */
+		char* acceptEncoding = NULL;
 
 		/* we might have options and then the url and the file */
 		/* let's process the options first */
@@ -206,6 +317,26 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 					theResult = TCL_ERROR;
 					break;
 				}
+			} else if (strcmp(theOption, "--append-http-header") == 0) {
+				/* check we also have the parameter */
+				if (optioncrsr < lastoption) {
+					optioncrsr++;
+					if (numHTTPHeaders < MAXHTTPHEADERS) {
+						httpHeaders[numHTTPHeaders++] = Tcl_GetString(objv[optioncrsr]);
+					} else {
+						Tcl_SetResult(interp,
+							"curl fetch: Too many --append-http-header options",
+							TCL_STATIC);
+						theResult = TCL_ERROR;
+						break;
+					}
+				} else {
+					Tcl_SetResult(interp,
+						"curl fetch: --append-http-header option requires a parameter",
+						TCL_STATIC);
+					theResult = TCL_ERROR;
+					break;
+				}
 			} else if (strcmp(theOption, "--progress") == 0) {
 				/* check we also have the parameter */
 				if (optioncrsr < lastoption) {
@@ -219,6 +350,8 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 					theResult = TCL_ERROR;
 					break;
 				}
+			} else if (strcmp(theOption, "--enable-compression") == 0) {
+				acceptEncoding = "";
 			} else {
 				Tcl_ResetResult(interp);
 				Tcl_AppendResult(interp, "curl fetch: unknown option ", theOption, NULL);
@@ -257,10 +390,29 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			break;
 		}
 
-		/* Create the CURL handle */
+#if LIBCURL_VERSION_NUM >= 0x074d00
+		cleanup_handle_if_needed(theURL);
+#endif
+
+		/* Create the CURL handles */
+		if (theMHandle == NULL) {
+			/* Re-use existing multi handle if theMHandle isn't NULL */
+			theMHandle = curl_multi_init();
+			if (theMHandle == NULL) {
+				theResult = TCL_ERROR;
+				Tcl_SetResult(interp, "error in curl_multi_init", TCL_STATIC);
+				break;
+			}
+		}
+
 		if (theHandle == NULL) {
 			/* Re-use existing handle if theHandle isn't NULL */
 			theHandle = curl_easy_init();
+			if (theHandle == NULL) {
+				theResult = TCL_ERROR;
+				Tcl_SetResult(interp, "error in curl_easy_init", TCL_STATIC);
+				break;
+			}
 		}
 		/* If we're re-using a handle, the previous call did ensure to reset it
 		 * to the default state using curl_easy_reset(3) */
@@ -415,8 +567,25 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			}
 		}
 
+		/* a CURLOPT_ACCEPT_ENCODING of "" means to let cURL write the
+		 * Accept-Encoding header for you, based on what the library
+		 * was compiled to support.
+		 * A value of NULL disables all attempts at decompressing responses.
+		*/
+#ifdef _CURL_ENCODING
+		theCurlCode = curl_easy_setopt(theHandle, _CURL_ENCODING, acceptEncoding);
+		if (theCurlCode != CURLE_OK) {
+			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
+			break;
+		}
+#endif
+
 		/* Clear the Pragma: no-cache header */
 		headers = curl_slist_append(headers, "Pragma:");
+		/* Append any optional headers */
+		for (int iH = 0; iH < numHTTPHeaders; ++iH) {
+			headers = curl_slist_append(headers, httpHeaders[iH]);
+		}
 		theCurlCode = curl_easy_setopt(theHandle, CURLOPT_HTTPHEADER, headers);
 		if (theCurlCode != CURLE_OK) {
 			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
@@ -429,61 +598,147 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			break;
 		}
 
-		/* actually fetch the resource */
-		theCurlCode = curl_easy_perform(theHandle);
-		if (theCurlCode != CURLE_OK) {
-			performFailed = true;
+		/* add the easy handle to the multi handle */
+		theCurlMCode = curl_multi_add_handle(theMHandle, theHandle);
+		if (theCurlMCode != CURLM_OK) {
+			theResult = SetResultFromCurlMErrorCode(interp, theCurlMCode);
 			break;
 		}
+		handleAdded = true;
+
+		/* select(2) the file descriptors used by curl and interleave with
+		 * checks for TclX signals */
+		do {
+			int rc; /* select() return code */
+
+			/* arguments for select(2) */
+			int nfds;
+			fd_set readfds;
+			fd_set writefds;
+			fd_set errorfds;
+			struct timeval timeout;
+
+			long curl_timeout = -1;
+
+			/* curl_multi_timeout introduced in libcurl 7.15.4 */
+#if LIBCURL_VERSION_NUM >= 0x070f04
+			/* get the next timeout */
+			theCurlMCode = curl_multi_timeout(theMHandle, &curl_timeout);
+			if (theCurlMCode != CURLM_OK) {
+				theResult = SetResultFromCurlMErrorCode(interp, theCurlMCode);
+				break;
+			}
+#endif
+
+			timeout.tv_sec = 1;
+			timeout.tv_usec = 0;
+			/* convert the timeout into a suitable format for select(2) and
+			 * limit the timeout to 1 second at most */
+			if (curl_timeout >= 0 && curl_timeout < 1000) {
+				timeout.tv_sec = 0;
+				/* convert ms to us */
+				timeout.tv_usec = curl_timeout * 1000;
+			}
+
+			/* get the fd sets for select(2) */
+			FD_ZERO(&readfds);
+			FD_ZERO(&writefds);
+			FD_ZERO(&errorfds);
+			theCurlMCode = curl_multi_fdset(theMHandle, &readfds, &writefds, &errorfds, &nfds);
+			if (theCurlMCode != CURLM_OK) {
+				theResult = SetResultFromCurlMErrorCode(interp, theCurlMCode);
+				break;
+			}
+
+			/* The value of nfds is guaranteed to be >= -1. Passing nfds + 1 to
+			 * select(2) makes the case of nfds == -1 a sleep. */
+			rc = select(nfds + 1, &readfds, &writefds, &errorfds, &timeout);
+			if (-1 == rc) {
+				/* check for signals first to avoid breaking our special
+				 * handling of SIGINT and SIGTERM */
+				if (Tcl_AsyncReady()) {
+					theResult = Tcl_AsyncInvoke(interp, theResult);
+					if (theResult != TCL_OK) {
+						break;
+					}
+				}
+
+				/* select error */
+				Tcl_SetResult(interp, strerror(errno), TCL_VOLATILE);
+				theResult = TCL_ERROR;
+				break;
+			}
+
+			/* timeout or activity */
+			theCurlMCode = curl_multi_perform(theMHandle, &running);
+
+			/* process signals from TclX */
+			if (Tcl_AsyncReady()) {
+				theResult = Tcl_AsyncInvoke(interp, theResult);
+				if (theResult != TCL_OK) {
+					break;
+				}
+			}
+		} while (running > 0);
+
+		/* Find out whether the transfer succeeded or failed. */
+		info = curl_multi_info_read(theMHandle, &running);
+		if (running > 0) {
+			fprintf(stderr, "Warning: curl_multi_info_read has %d more structs available\n", running);
+		}
+
+		/* free header memory */
+		curl_slist_free_all(headers);
 
 		/* signal cleanup to the progress callback */
 		if (noprogress == 0 && strcmp(progressCallback.proc, "builtin") != 0) {
 			CurlProgressCleanup(&progressCallback);
 		}
 
+		/* check for errors in the loop */
+		if (theResult != TCL_OK || theCurlMCode != CURLM_OK) {
+			break;
+		}
+
+		/* we should always get CURLMSG_DONE unless we aborted due to a Tcl
+		 * signal */
+		if (info == NULL) {
+			Tcl_SetResult(interp, "curl_multi_info_read() returned NULL", TCL_STATIC);
+			theResult = TCL_ERROR;
+			break;
+		}
+
+		if (info->msg != CURLMSG_DONE) {
+			snprintf(theErrorString, sizeof(theErrorString), "curl_multi_info_read() returned unexpected {.msg = %d, .data.result = %d}", info->msg, info->data.result);
+			Tcl_SetResult(interp, theErrorString, TCL_VOLATILE);
+			theResult = TCL_ERROR;
+			break;
+		}
+
+		if (info->data.result != CURLE_OK) {
+			/* execution failed, use the error string if it is set */
+			if (theErrorString[0] != '\0') {
+				Tcl_SetResult(interp, theErrorString, TCL_VOLATILE);
+			} else {
+				/* When the error buffer does not hold useful information,
+				 * generate our own message. Use a larger buffer since we add
+				 * a significant amount of text. */
+				char errbuf[256 + CURL_ERROR_SIZE];
+				snprintf(errbuf, sizeof(errbuf),
+					"curl_multi_info_read() returned {.msg = CURLMSG_DONE, "
+					".data.result = %d (!= CURLE_OK)}, but the error buffer "
+					"is not set. curl_easy_strerror(.data.result): %s",
+					info->data.result, curl_easy_strerror(info->data.result));
+				Tcl_SetResult(interp, errbuf, TCL_VOLATILE);
+			}
+			theResult = TCL_ERROR;
+			break;
+		}
+
 		/* close the file */
 		(void) fclose(theFile);
 		theFile = NULL;
 
-#if LIBCURL_VERSION_NUM == 0x070d01 /* work around broken Tiger version of cURL */
-		if (remotetime) {
-			FILE *fp;
-			char *tmp, *p;
-			char buf[BUFSIZ];
-			size_t size;
-
-			tmp = tmpnam(NULL);
-			fp = fopen(tmp, "w");
-			if (fp == NULL) {
-				Tcl_SetResult(interp, strerror(errno), TCL_VOLATILE);
-				theResult = TCL_ERROR;
-				break;
-			}
-			theFile = fopen(theFilePath, "r");
-			if (theFile == NULL) {
-				Tcl_SetResult(interp, strerror(errno), TCL_VOLATILE);
-				theResult = TCL_ERROR;
-				break;
-			}
-			if ((p = fgets(buf, BUFSIZ, theFile)) != NULL) {
-				/* skip stray header escaping into output */
-				if (strncmp(p, "Last-Modified:", 14) != 0)
-					rewind(theFile);
-			}
-			while ((size = fread(buf, 1, BUFSIZ, theFile)) > 0) {
-				fwrite(buf, 1, size, fp);
-			}
-			(void) fclose(theFile);
-			theFile = NULL;
-			fclose(fp);
-			if (rename(tmp, theFilePath) != 0) {
-				Tcl_SetResult(interp, strerror(errno), TCL_VOLATILE);
-				theResult = TCL_ERROR;
-				break;
-			}
-		}
-#endif
-
 		if (remotetime) {
 			theCurlCode = curl_easy_getinfo(theHandle, CURLINFO_FILETIME, &theFileTime);
 			if (theCurlCode == CURLE_OK && theFileTime > 0) {
@@ -494,21 +749,19 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			}
 		}
 
-		/* free header memory */
-		curl_slist_free_all(headers);
-
 		/* If --effective-url option was given, set given variable name to last effective url used by curl */
 		if (effectiveURLVarName != NULL) {
 			theCurlCode = curl_easy_getinfo(theHandle, CURLINFO_EFFECTIVE_URL, &effectiveURL);
 			Tcl_SetVar(interp, effectiveURLVarName,
-				(effectiveURL == NULL || theCurlCode != CURLE_OK) ? "" : effectiveURL,
-				0);
+				(effectiveURL == NULL || theCurlCode != CURLE_OK) ? "" : effectiveURL, 0);
 		}
 	} while (0);
 
-	if (performFailed) {
-		Tcl_SetResult(interp, theErrorString, TCL_VOLATILE);
-		theResult = TCL_ERROR;
+	if (handleAdded) {
+		/* Remove the handle from the multi handle, but ignore errors to avoid
+		 * cluttering the real error info that might be somewhere further up */
+		curl_multi_remove_handle(theMHandle, theHandle);
+		handleAdded = false;
 	}
 
 	/* reset the connection */
@@ -530,10 +783,9 @@ CurlFetchCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
  * @param objv			parameters
  */
 int
-CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
-	CURL* theHandle = NULL;
 	FILE* theFile = NULL;
 
 	do {
@@ -545,6 +797,7 @@ CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		CURLcode theCurlCode;
 		long theModDate;
 		long userModDate;
+		const char* userAgent = PACKAGE_NAME "/" PACKAGE_VERSION " libcurl/" LIBCURL_VERSION;
 
 		optioncrsr = 2;
 		lastoption = objc - 3;
@@ -615,6 +868,13 @@ CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			break;
 		}
 
+		/* -A option */
+		theCurlCode = curl_easy_setopt(theHandle, CURLOPT_USERAGENT, userAgent);
+		if (theCurlCode != CURLE_OK) {
+			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
+			break;
+		}
+
 		/* --max-redirs option, same default as curl command line */
 		theCurlCode = curl_easy_setopt(theHandle, CURLOPT_MAXREDIRS, 50);
 		if (theCurlCode != CURLE_OK) {
@@ -767,10 +1027,9 @@ CurlIsNewerCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
  * @param objv			parameters
  */
 int
-CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
-	CURL* theHandle = NULL;
 	FILE* theFile = NULL;
 
 	do {
@@ -779,6 +1038,7 @@ CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		int ignoresslcert = 0;
 		char theSizeString[32];
 		const char* theURL;
+		const char* theUserPassString = NULL;
 		CURLcode theCurlCode;
 		double theFileSize;
 
@@ -790,6 +1050,18 @@ CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 
 			if (strcmp(theOption, "--ignore-ssl-cert") == 0) {
 				ignoresslcert = 1;
+			} else if (strcmp(theOption, "-u") == 0) {
+				/* check we also have the parameter */
+				if (optioncrsr < lastoption) {
+					optioncrsr++;
+					theUserPassString = Tcl_GetString(objv[optioncrsr]);
+				} else {
+					Tcl_SetResult(interp,
+						"curl getsize: -u option requires a parameter",
+						TCL_STATIC);
+					theResult = TCL_ERROR;
+					break;
+				}
 			} else {
 				Tcl_ResetResult(interp);
 				Tcl_AppendResult(interp, "curl getsize: unknown option ", theOption, NULL);
@@ -806,7 +1078,7 @@ CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 		}
 
 		/* first (second) parameter is the url */
-		if (objc < 3 || objc > 4) {
+		if (objc < 3 || objc > 6) {
 			Tcl_WrongNumArgs(interp, 1, objv, "getsize [--ignore-ssl-cert] url");
 			theResult = TCL_ERROR;
 			break;
@@ -908,6 +1180,15 @@ CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 			}
 		}
 
+        /* set the l/p, if any */
+		if (theUserPassString) {
+			theCurlCode = curl_easy_setopt(theHandle, CURLOPT_USERPWD, theUserPassString);
+			if (theCurlCode != CURLE_OK) {
+				theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
+				break;
+			}
+		}
+
 		/* skip the header data */
 		theCurlCode = curl_easy_setopt(theHandle, CURLOPT_HEADER, 0);
 		if (theCurlCode != CURLE_OK) {
@@ -976,10 +1257,9 @@ CurlGetSizeCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
  * @param objv			parameters
  */
 int
-CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
+CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
 {
 	int theResult = TCL_OK;
-	CURL* theHandle = NULL;
 	FILE* theFile = NULL;
 
 	do {
@@ -1184,16 +1464,17 @@ CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 
 		/* actually perform the POST */
 		theCurlCode = curl_easy_perform(theHandle);
-		if (theCurlCode != CURLE_OK) {
-			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
-			break;
-		}
 
 		/* signal cleanup to the progress callback */
 		if (noprogress == 0 && strcmp(progressCallback.proc, "builtin") != 0) {
 			CurlProgressCleanup(&progressCallback);
 		}
 
+		if (theCurlCode != CURLE_OK) {
+			theResult = SetResultFromCurlErrorCode(interp, theCurlCode);
+			break;
+		}
+
 		/* close the file */
 		(void) fclose(theFile);
 		theFile = NULL;
@@ -1211,6 +1492,75 @@ CurlPostCmd(Tcl_Interp* interp, int objc, Tcl_Obj* CONST objv[])
 	return theResult;
 }
 
+/* Fill in global libcurl_version_info structure */
+static void set_curl_version_info(void)
+{
+    if (libcurl_version_info == NULL) {
+        libcurl_version_info = curl_version_info(CURLVERSION_NOW);
+    }
+}
+
+/**
+ * curl version subcommand entry point.
+ *
+ * @param interp		current interpreter
+ * @param objc			number of parameters
+ * @param objv			parameters
+ */
+int
+CurlVersionCmd(Tcl_Interp* interp, int objc, Tcl_Obj* const objv[])
+{
+	int theResult = TCL_OK;
+	Tcl_Obj *tcl_result = NULL;
+	curl_version_info_data *theVersionInfo;
+
+    set_curl_version_info();
+    if (libcurl_version_info == NULL) {
+        return TCL_ERROR;
+    }
+    theVersionInfo = libcurl_version_info;
+
+	tcl_result = Tcl_NewDictObj();
+
+	// info from the curl version we were built against:
+	Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("VERSION_NOW-build", -1), Tcl_NewIntObj(CURLVERSION_NOW));
+	Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("VERSION-build", -1), Tcl_NewStringObj(LIBCURL_VERSION, -1));
+	// runtime info from the libcurl we are actually using:
+	Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("VERSION-runtime", -1), Tcl_NewStringObj(theVersionInfo->version, -1));
+	Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("HOST", -1), Tcl_NewStringObj(theVersionInfo->host, -1));
+	Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("SSL_BACKEND", -1), Tcl_NewStringObj(theVersionInfo->ssl_version, -1));
+	Tcl_Obj *tcl_prots = Tcl_NewListObj(0, NULL);
+	const char * const *prot = theVersionInfo->protocols;
+	while (*prot) {
+		Tcl_ListObjAppendElement(interp, tcl_prots, Tcl_NewStringObj(*prot, -1));
+		prot++;
+	}
+	Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("PROTOCOLS", -1), tcl_prots);
+// returning feature names requires a build-time and a runtime check:
+#if LIBCURL_VERSION_NUM >= 0x078700
+	if (theVersionInfo->age >= CURLVERSION_ELEVENTH && theVersionInfo->feature_names) {
+		Tcl_Obj *tcl_feats = Tcl_NewListObj(0, NULL);
+		const char * const *feats = theVersionInfo->feature_names;
+		while (*feats) {
+			Tcl_ListObjAppendElement(interp, tcl_feats, Tcl_NewStringObj(*feats, -1));
+			feats++;
+		}
+		Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("FEATURES", -1), tcl_feats);
+	} else
+#endif
+	{
+		// it would be possible to generate a list of feature names from the feature mask
+		// but tedious because it would involve checking for the existence of many of the 
+		// feature tokens; probably not justified for the intended use of the `curl version`
+		// command (primarily debugging).
+		Tcl_DictObjPut(interp, tcl_result, Tcl_NewStringObj("FEATURE_MASK", -1), Tcl_NewIntObj(theVersionInfo->features));
+	}
+
+	Tcl_SetObjResult(interp, tcl_result);
+
+	return theResult;
+}
+
 /**
  * curl command entry point.
  *
@@ -1224,17 +1574,18 @@ CurlCmd(
 		ClientData clientData UNUSED,
 		Tcl_Interp* interp,
 		int objc,
-		Tcl_Obj* CONST objv[])
+		Tcl_Obj* const objv[])
 {
 	typedef enum {
 		kCurlFetch,
 		kCurlIsNewer,
 		kCurlGetSize,
-		kCurlPost
+		kCurlPost,
+		kCurlVersion
 	} EOption;
 
 	static const char *options[] = {
-		"fetch", "isnewer", "getsize", "post", NULL
+		"fetch", "isnewer", "getsize", "post", "version", NULL
 	};
 	int theResult = TCL_OK;
 	EOption theOptionIndex;
@@ -1243,18 +1594,18 @@ CurlCmd(
 	/* TODO: use dispatch_once when we drop Leopard support */
 	pthread_once(&once, CurlInit);
 
-	if (objc < 3) {
-		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
-		return TCL_ERROR;
-	}
-
-	theResult = Tcl_GetIndexFromObj(
+	theResult = objc > 1 ? Tcl_GetIndexFromObj(
 				interp,
 				objv[1],
 				options,
 				"option",
 				0,
-				(int*) &theOptionIndex);
+				(int*) &theOptionIndex) : TCL_ERROR;
+	if (theResult != TCL_OK || (objc < 3 && theOptionIndex != kCurlVersion)) {
+		Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
+		return TCL_ERROR;
+	}
+
 	if (theResult == TCL_OK) {
 		switch (theOptionIndex) {
 		case kCurlFetch:
@@ -1269,6 +1620,9 @@ CurlCmd(
 		case kCurlPost:
 			theResult = CurlPostCmd(interp, objc, objv);
 			break;
+		case kCurlVersion:
+			theResult = CurlVersionCmd(interp, objc, objv);
+			break;
 		}
 	}
 
@@ -1280,7 +1634,7 @@ CurlCmd(
  * libcurl will never be cleaned (where should I plug the hook?)
  */
 void
-CurlInit()
+CurlInit(void)
 {
 	curl_global_init(CURL_GLOBAL_ALL);
 }
@@ -1404,6 +1758,7 @@ static void CurlProgressCleanup(
 	/*
 	 * Transfer complete, signal the progress callback
 	 */
+	Tcl_InterpState state;
 
 	/*
 	 * Command string, a space followed "finish" plus the trailing \0.
@@ -1418,5 +1773,9 @@ static void CurlProgressCleanup(
 		abort();
 	}
 
+	/* make sure to save and restore the interpreter state so a potential error
+	 * message doesn't get lost */
+	state = Tcl_SaveInterpState(callback->interp, 0);
 	Tcl_EvalEx(callback->interp, commandBuffer, len, TCL_EVAL_GLOBAL);
+	Tcl_RestoreInterpState(callback->interp, state);
 }
