@@ -1,8 +1,8 @@
 # et:ts=4
 # portpatch.tcl
-# $Id: portpatch.tcl 69336 2010-07-01 01:32:50Z jmr@macports.org $
 #
-# Copyright (c) 2002 - 2003 Apple Computer, Inc.
+# Copyright (c) 2004, 2006-2007, 2009-2011 The MacPorts Project
+# Copyright (c) 2002 - 2003 Apple Inc.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -13,7 +13,7 @@
 # 2. Redistributions in binary form must reproduce the above copyright
 #    notice, this list of conditions and the following disclaimer in the
 #    documentation and/or other materials provided with the distribution.
-# 3. Neither the name of Apple Computer, Inc. nor the names of its contributors
+# 3. Neither the name of Apple Inc. nor the names of its contributors
 #    may be used to endorse or promote products derived from this software
 #    without specific prior written permission.
 #
@@ -50,7 +50,7 @@ options patch.asroot
 default patch.asroot no
 default patch.dir {${worksrcpath}}
 default patch.cmd {[portpatch::build_getpatchtype]}
-default patch.pre_args -p0
+default patch.pre_args {-t -N -p0}
 
 proc portpatch::build_getpatchtype {args} {
     if {![exists patch.type]} {
@@ -68,14 +68,14 @@ proc portpatch::build_getpatchtype {args} {
 }
 
 proc portpatch::patch_main {args} {
-    global UI_PREFIX usealtworkpath altprefix
+    global UI_PREFIX
 
     # First make sure that patchfiles exists and isn't stubbed out.
-    if {![exists patchfiles] || [option patchfiles] == ""} {
+    if {![exists patchfiles] || [option patchfiles] eq ""} {
         return 0
     }
 
-    ui_msg "$UI_PREFIX [format [msgcat::mc "Applying patches to %s"] [option name]]"
+    ui_notice "$UI_PREFIX [format [msgcat::mc "Applying patches to %s"] [option subport]]"
 
     foreach patch [option patchfiles] {
         set patch_file [getdistname $patch]
@@ -83,8 +83,6 @@ proc portpatch::patch_main {args} {
             lappend patchlist [option filespath]/$patch_file
         } elseif {[file exists [option distpath]/$patch_file]} {
             lappend patchlist [option distpath]/$patch_file
-        } elseif {!$usealtworkpath && [file exists "${altprefix}[option distpath]/$patch_file"]} {
-            lappend patchlist "${altprefix}[option distpath]/$patch_file"
         } else {
             return -code error [format [msgcat::mc "Patch file %s is missing"] $patch]
         }
@@ -95,12 +93,20 @@ proc portpatch::patch_main {args} {
 
     set gzcat "[findBinary gzip $portutil::autoconf::gzip_path] -dc"
     set bzcat "[findBinary bzip2 $portutil::autoconf::bzip2_path] -dc"
+    catch {set xzcat "[findBinary xz $portutil::autoconf::xz_path] -dc"}
+
     foreach patch $patchlist {
-        ui_info "$UI_PREFIX [format [msgcat::mc "Applying %s"] $patch]"
+        ui_info "$UI_PREFIX [format [msgcat::mc "Applying %s"] [file tail $patch]]"
         switch -- [file extension $patch] {
             .Z -
             .gz {command_exec patch "$gzcat \"$patch\" | (" ")"}
             .bz2 {command_exec patch "$bzcat \"$patch\" | (" ")"}
+            .xz {
+                if {[info exists xzcat]} {
+                    command_exec patch "$xzcat \"$patch\" | (" ")"
+                } else {
+                    return -code error [msgcat::mc "xz binary not found; port needs to add 'depends_patch bin:xz:xz'"]
+                }}
             default {command_exec patch "" "< '$patch'"}
         }
     }
